// RunCPM extensions
// DS3231 module connected via I2C bus

#ifndef DATETIME_H
#define DATETIME_H
#define DS3231_I2C_ADDRESS 0x68


// Convert normal decimal number to binary coded decimal
uint8 decToBcd(uint8 val)
{
	return (val / 10 * 16) + (val % 10);
}


// Convert binary coded decimal to normal decimal number
uint8 bcdToDec(uint8 val)
{
	return (val / 16 * 10) + (val % 16);
}


// Convert binary coded decimal from memory to normal decimal number
uint8 addrToDec (uint16 addr)
{
	return bcdToDec(_RamRead(addr));
}


// Convert binary coded decimal year value from memory to normal decimal number
uint16 addrToDecY (uint16 addr)
{
	uint8 year = addrToDec(addr);
	return year > 77 ? year + 1900 : year + 2000;
}


// User date/time callback function.
void dateTime(uint16_t* date, uint16_t* time)
{
	uint16 year;
	uint8 month, day, hour, minute, second;

	Wire.beginTransmission(DS3231_I2C_ADDRESS);
	Wire.write(0);  // set DS3231 register pointer to 0
	Wire.endTransmission();
	Wire.requestFrom(DS3231_I2C_ADDRESS, 7);
	// request 7 bytes of data from DS3231 starting from register 0
	second = bcdToDec(Wire.read());
	minute = bcdToDec(Wire.read());
	hour   = bcdToDec(Wire.read());
	Wire.read(); // day of week - not used
	day    = bcdToDec(Wire.read());
	month  = bcdToDec(Wire.read());
	year   = bcdToDec(Wire.read()) + 2000;

	*date = FAT_DATE(year, month, day);
	*time = FAT_TIME(hour, minute, second);
}


// Get file date/time to DMA buffer
uint8 _GetFileStamp(uint16 fcbaddr, uint16 addr)
{
	CPM_FCB* F = (CPM_FCB*)_RamSysAddr(fcbaddr);
	uint8 result = 0xff;

	if (!_SelectDisk(F->dr)) {
		dir_t dir;
		_FCBtoHostname(fcbaddr, &filename[0]);
		File f = SD.open((char*)filename, O_READ);
		f.dirEntry(&dir);
		_RamWrite(addr,      decToBcd(FAT_YEAR(dir.creationDate) % 100));
		_RamWrite(addr +  1, decToBcd(FAT_MONTH(dir.creationDate)));
		_RamWrite(addr +  2, decToBcd(FAT_DAY(dir.creationDate)));
		_RamWrite(addr +  3, decToBcd(FAT_HOUR(dir.creationTime)));
		_RamWrite(addr +  4, decToBcd(FAT_MINUTE(dir.creationTime)));
		_RamWrite(addr +  5, decToBcd(FAT_YEAR(dir.lastAccessDate) % 100));
		_RamWrite(addr +  6, decToBcd(FAT_MONTH(dir.lastAccessDate)));
		_RamWrite(addr +  7, decToBcd(FAT_DAY(dir.lastAccessDate)));
		_RamWrite(addr +  8, 0); // dir.lastAccessTime doesn't exist
		_RamWrite(addr +  9, 0);
		_RamWrite(addr + 10, decToBcd(FAT_YEAR(dir.lastWriteDate) % 100));
		_RamWrite(addr + 11, decToBcd(FAT_MONTH(dir.lastWriteDate)));
		_RamWrite(addr + 12, decToBcd(FAT_DAY(dir.lastWriteDate)));
		_RamWrite(addr + 13, decToBcd(FAT_HOUR(dir.lastWriteTime)));
		_RamWrite(addr + 14, decToBcd(FAT_MINUTE(dir.lastWriteTime)));
		f.close();
		result = 0x01;
	}
	return result;
}


// Set file date/time from DMA buffer
uint8 _SetFileStamp(uint16 fcbaddr, uint16 addr)
{
	CPM_FCB* F = (CPM_FCB*)_RamSysAddr(fcbaddr);
	uint8 result = 0xff;

	if (!_SelectDisk(F->dr)) {
		SdFile::dateTimeCallbackCancel();  // disable date/time callback function
		_FCBtoHostname(fcbaddr, &filename[0]);
		File f = SD.open((char*)filename, O_CREAT | O_WRITE);
		// set creation date time
		f.timestamp(T_CREATE,
			addrToDecY(addr),    // year
			addrToDec(addr + 1), // month
			addrToDec(addr + 2), // day
			addrToDec(addr + 3), // hour
			addrToDec(addr + 4), // minute
			0);                  // second - not used
		// set access date time
		f.timestamp(T_ACCESS,
			addrToDecY(addr + 5), // year
			addrToDec(addr + 6),  // month
			addrToDec(addr + 7),  // day
			addrToDec(addr + 8),  // hour
			addrToDec(addr + 9),  // minute
			0);                   // second - not used
		// set write/modification date time
		f.timestamp(T_WRITE,
			addrToDecY(addr + 10), // year
			addrToDec(addr + 11),  // month
			addrToDec(addr + 12),  // day
			addrToDec(addr + 13),  // hour
			addrToDec(addr + 14),  // minute
			0);                    // second - not used
		f.close();
		result = 0x01;
		SdFile::dateTimeCallback(dateTime); // enable date/time callback function
	}
	return result;
}


// Get system date/time
uint8 _GetTime(uint16 addr)
{
	Wire.beginTransmission(DS3231_I2C_ADDRESS);
	Wire.write(0);   // set DS3231 register pointer to 0
	if (Wire.endTransmission()) {
		return 0xff;  // error
	}
	// request seven bytes of data from DS3231 starting register 0
	Wire.requestFrom(DS3231_I2C_ADDRESS, 7);
	for (int i = 5; i >= 0; i--) {
		_RamWrite(addr + i, Wire.read());
		if (i == 3) {
			Wire.read(); // day of week - not used
		}
	}
	return 0x01;  // no error
}


// Set system date/time
uint8 _SetTime(uint16 addr)
{
	Wire.beginTransmission(DS3231_I2C_ADDRESS);
	Wire.write(0);   // set DS3231 register pointer to 0
	// write bytes of data to DS3231 starting register 0
	for (int i = 5; i >= 0; i--) {
		Wire.write(_RamRead(addr + i));
		if (i == 3) {
			Wire.write(1); // day of week not used
		}
	}
	if (Wire.endTransmission()) {
		return 0xff;  // error
	} else {
		return 0x01;  // no error
	}
}


// ----- EEPROM-------------------------

#define AT24C32_I2C_ADDRESS 0x57

// Write value into EEPROM
void writeEEPROM(uint16 addr, uint8 data)
{
	Wire.beginTransmission(AT24C32_I2C_ADDRESS);
	Wire.write(addr >> 8);   // MSB
	Wire.write(addr & 0xff); // LSB
	Wire.write(data);
	Wire.endTransmission();
	delay(5);
}


// Read value from EEPROM
uint8 readEEPROM(uint16 addr)
{
	uint8 data = 0xff;
	Wire.beginTransmission(AT24C32_I2C_ADDRESS);
	Wire.write(addr >> 8);   // MSB
	Wire.write(addr & 0xff); // LSB
	Wire.endTransmission();
	Wire.requestFrom(AT24C32_I2C_ADDRESS, 1);
	if (Wire.available()) {
		data = Wire.read();
	}
	return data;
}


// ----- HW SETTINGS -------------------
/*
Baudrate  EEPROM
   300     0x01
   600     0x02
  1200     0x04
  2400     0x08
  4800     0x10
  9600     0x20 (the default)
 14400     0x30
 19200     0x40
 28800     0x60
 38400     0x80
 57600     0xC0
*/

uint16 serial_param[] = {
	UARTClass::Mode_8N1,	// 0x00 (the default)
	UARTClass::Mode_8E1,	// 0x01
	UARTClass::Mode_8O1,	// 0x02
	UARTClass::Mode_8M1,	// 0x03
	UARTClass::Mode_8S1,	// 0x04
	USARTClass::Mode_5N1,	// 0x05
	USARTClass::Mode_6N1,	// 0x06
	USARTClass::Mode_7N1,	// 0x07
	USARTClass::Mode_5N2,	// 0x08
	USARTClass::Mode_6N2,	// 0x09
	USARTClass::Mode_7N2,	// 0x0a
	USARTClass::Mode_8N2,	// 0x0b
	USARTClass::Mode_5E1,	// 0x0c
	USARTClass::Mode_6E1,	// 0x0d
	USARTClass::Mode_7E1,	// 0x0e
	USARTClass::Mode_5E2,	// 0x0f
	USARTClass::Mode_6E2,	// 0x10
	USARTClass::Mode_7E2,	// 0x11
	USARTClass::Mode_8E2,	// 0x12
	USARTClass::Mode_5O1,	// 0x13
	USARTClass::Mode_6O1,	// 0x14
	USARTClass::Mode_7O1,	// 0x15
	USARTClass::Mode_5O2,	// 0x16
	USARTClass::Mode_6O2,	// 0x17
	USARTClass::Mode_7O2,	// 0x18
	USARTClass::Mode_8O2,	// 0x19
	USARTClass::Mode_5M1,	// 0x1a
	USARTClass::Mode_6M1,	// 0x1b
	USARTClass::Mode_7M1,	// 0x1c
	USARTClass::Mode_5M2,	// 0x1d
	USARTClass::Mode_6M2,	// 0x1e
	USARTClass::Mode_7M2,	// 0x1f
	USARTClass::Mode_8M2,	// 0x20
	USARTClass::Mode_5S1,	// 0x21
	USARTClass::Mode_6S1,	// 0x22
	USARTClass::Mode_7S1,	// 0x23
	USARTClass::Mode_5S2,	// 0x24
	USARTClass::Mode_6S2,	// 0x25
	USARTClass::Mode_7S2,	// 0x26
	USARTClass::Mode_8S2	// 0x27
};


void set_HW()
{
	pinMode(LED, OUTPUT);
	digitalWrite(LED, LOW);
	pinMode(CONSPIN, INPUT_PULLUP);
	_RamWrite(IObyte, digitalRead(CONSPIN) ? readEEPROM(8) : readEEPROM(9));
	_initlst();
	// serial ports init
	uint8 i = readEEPROM(1);
	if (i < 5) {
		Serial.begin(readEEPROM(0) * 300, (UARTClass::UARTModes)serial_param[i]);
	}
	i = readEEPROM(3);
	if (i < 40) {
		Serial1.begin(readEEPROM(2) * 300, (USARTClass::USARTModes)serial_param[i]);
	}
	i = readEEPROM(5);
	if (i < 40) {
		Serial2.begin(readEEPROM(4) * 300, (USARTClass::USARTModes)serial_param[i]);
	}
	i = readEEPROM(7);
	if (i < 40) {
		Serial3.begin(readEEPROM(6) * 300, (USARTClass::USARTModes)serial_param[i]);
	}
}


// parallel port busy -> true
bool getListSt()
{
	return digitalRead(busy) == HIGH;
}

#endif

